/*
UTMapFlagger - A small app to set flags on Unreal Tournament Maps
Copyright (C) 2004  Malone

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#define UTMAPFLAGGERVERSION "1.0"

#include <stdio.h>
#include <sys/stat.h>
#include "getopt.h"

char* progName = NULL;

#define ALLOW_DOWNLOAD 0x01
#define CLIENT_OPTIONAL 0x02
#define SERVERSIDE_ONLY 0x04

#define UNSECURE 0x10
#define OFFICIAL 0x20

#define FLAG_OFFSET 8

void displayHelp(){
	printf("UTMapFlagger %s by Malone\n",UTMAPFLAGGERVERSION);
	printf(" A small app to set flags on Unreal Tournament Maps\n");
	printf(" (tested on UT2K4 .ut2 maps)\n\n");

	printf("Usage: %s [-f] [--<flag>[=on|=off]] <mapfile>\n\n",progName);
	printf("  -f --force   Modify even if file isnt a real map\n\n");

	printf("  Available Flags:\n");
	printf("   --AllowDownload\n");
	printf("   --ClientOptional\n");
	printf("   --ServerSideOnly\n");
	printf("   --Unsecure\n");
	printf("   --Official\n\n");

	printf("Example: %s --Official=off mapname.ut2\n",progName);
	printf(" (removes the Official flag if it is set)\n");
}

int main(int argc, char **argv) {
	progName = argv[0];

	if(argc < 2){
		displayHelp();
		return 1;
	}

	int modify = 0;
	int c;
	char *mapfile = NULL;
	char cflags, nflags;
	FILE *mfile;

	char onMask = 0;
	char offMask = 0;
	char toggleMask = 0;


	static struct option long_options[] = {
		{"AllowDownload", 2, 0, 'a'},
		{"ClientOptional", 2, 0, 'c'},
		{"ServerSideOnly", 2, 0, 's'},
		{"Unsecure", 2, 0, 'u'},
		{"Official", 2, 0, 'o'},
		{"force", 0, 0, 'f'},
		{0, 0, 0, 0}
	};

	int errors = 0;
	int optchar;
	while ((optchar = getopt_long(argc, argv, "a::c::s::u::o::f", long_options, 0)) != -1){
		switch (optchar) {

			case 'a':
				if(optarg){
					if(strcmp(optarg,"on") == 0){
						onMask |= ALLOW_DOWNLOAD;
					}else if(strcmp(optarg,"off") == 0){
						offMask |= ALLOW_DOWNLOAD;
					}else{
						fprintf (stderr, "%s: `%s' is not a valid parameter for --AllowDownload (expecting 'on' or 'off')\n",progName,optarg);
						++errors;
					}
				}else{
					toggleMask |= ALLOW_DOWNLOAD;
				}
			break;

			case 'c':
				if(optarg){
					if(strcmp(optarg,"on") == 0){
						onMask |= CLIENT_OPTIONAL;
					}else if(strcmp(optarg,"off") == 0){
						offMask |= CLIENT_OPTIONAL;
					}else{
						fprintf (stderr, "%s: `%s' is not a valid parameter for --ClientOptional (expecting 'on' or 'off')\n",progName,optarg);
						++errors;
					}
				}else{
					toggleMask |= CLIENT_OPTIONAL;
				}
			break;

			case 's':
				if(optarg){
					if(strcmp(optarg,"on") == 0){
						onMask |= SERVERSIDE_ONLY;
					}else if(strcmp(optarg,"off") == 0){
						offMask |= SERVERSIDE_ONLY;
					}else{
						fprintf (stderr, "%s: `%s' is not a valid parameter for --ServerSideOnly (expecting 'on' or 'off')\n",progName,optarg);
						++errors;
					}
				}else{
					toggleMask |= SERVERSIDE_ONLY;
				}
			break;

			case 'u':
				if(optarg){
					if(strcmp(optarg,"on") == 0){
						onMask |= UNSECURE;
					}else if(strcmp(optarg,"off") == 0){
						offMask |= UNSECURE;
					}else{
						fprintf (stderr, "%s: `%s' is not a valid parameter for --Unsecure (expecting 'on' or 'off')\n",progName,optarg);
						++errors;
					}
				}else{
					toggleMask |= UNSECURE;
				}
			break;

			case 'o':
				if(optarg){
					if(strcmp(optarg,"on") == 0){
						onMask |= OFFICIAL;
					}else if(strcmp(optarg,"off") == 0){
						offMask |= OFFICIAL;
					}else{
						fprintf (stderr, "%s: `%s' is not a valid parameter for --Official (expecting 'on' or 'off')\n",progName,optarg);
						++errors;
					}
				}else{
					toggleMask |= OFFICIAL;
				}
			break;

			case 'f':
				modify = 1;
			break;

			default:
				++errors;
			break;
		} //switch
	}

	//Check for filename
	if (optind < argc) {
		mapfile = argv[optind];
		optind++;
	}else{
		fprintf (stderr, "%s: Please give the map filename ",progName);
	}

    if (optind < argc) {
		fprintf (stderr, "%s: Unknown options: ",progName);
		while (optind < argc) {
			fprintf (stderr, "\"%s\" ",argv[optind]);
			optind++;
		}
		fprintf (stderr, "\n");
		++errors;
    }

	if(errors > 0){
		//Crazy options, show usage and exit
		displayHelp();
		return 1;
	}


	struct stat fileinfo;
	if (stat(mapfile, &fileinfo) == 0){
		if(fileinfo.st_size < FLAG_OFFSET + 1){
      	fprintf(stderr,"%s: Map file too small\n",progName);
			return 1;
		}
   }else{
		fprintf(stderr,"%s: Couldnt get info on map file\n",progName);
		return 1;
	}

	if((mfile = fopen(mapfile, "r")) == NULL){
		fprintf(stderr,"%s: Couldnt open file for reading\n",progName);
		return 1;
	}

	//check the first byte of the file as a simple check for a valid map file
	if(getc(mfile) == 0xc1){
		modify = 1;
	}else{
		printf("[Warning]: file may not be a valid map!\n\n");
	}

	fseek(mfile, FLAG_OFFSET, SEEK_SET);
	cflags = getc(mfile);
	fclose(mfile);

	printf("Current Flags [");
	for(c = 128; c > 0; c = c >> 1){
		if(cflags & c){
			printf("1");
		}else{
			printf("0");
		}
	}
	printf("]:\n");

	if(cflags & ALLOW_DOWNLOAD){
		printf(" AllowDownload\n");
	}

	if(cflags & CLIENT_OPTIONAL){
		printf(" ClientOptional\n");
	}

	if(cflags & SERVERSIDE_ONLY){
		printf(" ServerSideOnly\n");
	}

	if(cflags & UNSECURE){
		printf(" Unsecure\n");
	}

	if(cflags & OFFICIAL){
		printf(" Official\n");
	}
	printf("\n");

	if(onMask > 0 || offMask > 0 || toggleMask > 0){
		nflags = cflags;
		nflags |= onMask;
		nflags &= ~(offMask);
		nflags ^= toggleMask;

		printf("    New Flags [");
		for(c = 128; c > 0; c = c >> 1){
			if(nflags & c){
				printf("1");
			}else{
				printf("0");
			}
		}
		printf("]:\n");

		if(nflags & ALLOW_DOWNLOAD){
			printf(" AllowDownload\n");
		}

		if(nflags & CLIENT_OPTIONAL){
			printf(" ClientOptional\n");
		}

		if(nflags & SERVERSIDE_ONLY){
			printf(" ServerSideOnly\n");
		}

		if(nflags & UNSECURE){
			printf(" Unsecure\n");
		}

		if(nflags & OFFICIAL){
			printf(" Official\n");
		}
		printf("\n");

		if(cflags == nflags){
			printf("No change to make\n");
			return 0;
		}

		if(modify){
			if((mfile = fopen(mapfile, "rb+")) == NULL){
				fprintf(stderr,"%s: Couldnt open file for writing\n",progName);
				return 1;
			}

			fseek(mfile, FLAG_OFFSET, SEEK_SET);
			cflags = putc(nflags,mfile);
			fclose(mfile);
			printf("Map file updated\n");
		}else{
			printf("Not writing to file. Use --force option to force modificaton\n");
		}
	}

	return 0;

}


